﻿using System;
using System.IO;
namespace ALchemyPatch
{
    class Program
    {
        static bool SigCompare(byte[] data, int offset, byte[] pattern, bool[] wildcards)
        {
            for(int i = 0; i < pattern.Length; i++)
            {
                if (wildcards[i])
                    continue;

                if (data[i + offset] != pattern[i])
                    return false;
            }
            return true;
        }

        static int SigSearch(byte[] data, int offset, string signature)
        {
            // convert string pattern to byte & wildcard arrays
            var cleanSig = signature.Replace(" ", "").Replace("0x", "");
            byte[] pattern = new byte[cleanSig.Length / 2];
            bool[] wildcards = new bool[cleanSig.Length / 2];
            for(int i = 0; i < cleanSig.Length/2; i++)
            {
                var byt = cleanSig.Substring(i*2, 2);
                if(byt == "??")
                {
                    pattern[i] = 0;
                    wildcards[i] = true;
                }
                else
                {
                    pattern[i] = Convert.ToByte(byt, 16);
                    wildcards[i] = false;
                }
            }

            // search!
            for(int i = offset; i < data.Length; i++)
            {
                if (SigCompare(data, i, pattern, wildcards))
                    return i;
            }

            return -1;
        }

        static bool PatchALchemyFile(string filePath)
        {
            if (!File.Exists(filePath))
            {
                Console.WriteLine($"Failed: couldn't find file '{filePath}' to patch...");
                return false;
            }

            Console.WriteLine($"Patching {filePath}...");
            var fileData = File.ReadAllBytes(filePath);

            // Find some code that's near the start of restrict func, in both exe & dll
            string restrictFuncSignature = "8B B4 24 ?? ?? ?? ?? 57 68 ?? ?? ?? ?? 56 E8 ?? ?? ?? ??";
            int restrictFuncAddr = SigSearch(fileData, 0, restrictFuncSignature);
            if (restrictFuncAddr < 0)
            {
                Console.WriteLine("  Failed: couldn't find restriction func!");
                return false;
            }

            Console.WriteLine($"  Restriction func at ~0x{restrictFuncAddr:X}");

            // Now find start of the function
            string restrictFuncStartSignature = "B8 ?? ?? ?? ?? E8";
            restrictFuncAddr = SigSearch(fileData, restrictFuncAddr - 0x40, restrictFuncStartSignature);
            if (restrictFuncAddr < 0)
            {
                Console.WriteLine("  Failed: couldn't find func start!");
                return false;
            }

            Console.WriteLine($"  Func start at 0x{restrictFuncAddr:X}");

            // Backup original...
            File.WriteAllBytes(filePath + ".bk", fileData);

            // Replace func with equivalent of "return 1";
            fileData[restrictFuncAddr + 0] = 0xB8;
            fileData[restrictFuncAddr + 1] = 0x01;
            fileData[restrictFuncAddr + 2] = 0x00;
            fileData[restrictFuncAddr + 3] = 0x00;
            fileData[restrictFuncAddr + 4] = 0x00;
            fileData[restrictFuncAddr + 5] = 0xC3;

            // Write patched exe
            File.WriteAllBytes(filePath, fileData);
            Console.WriteLine("Patch successful!");
            return true;
        }


        static void Main(string[] args)
        {
            if (PatchALchemyFile(@"ALchemy.exe"))
                PatchALchemyFile(@"dsound.dll");

            Console.WriteLine();
            Console.WriteLine("Press any key to exit.");
            Console.ReadKey();
        }
    }
}
